@@ -86,8 +86,8 @@ static const ThaiShapeTable Mac_shape_table = {
   { 0x93, 0x83, 0x84, 0x85, 0x86, 0x87, 0x8F, 0xEE },
   { 0x93, 0x98, 0x99, 0x9A, 0x9B, 0x9C, 0x8F, 0xEE },
   { 0x92, 0x00, 0x00, 0x94, 0x95, 0x96, 0x97 },
-  { 0xD8, 0xD9, 0xDA },
-  { 0xAD, 0x00, 0x00, 0xB0 },
+  { 0xFC, 0xFD, 0xFE },
+  { 0x90, 0x00, 0x00, 0x80 },
   { 0xED, 0xD2 }
 };
 
