@@ -26,14 +26,9 @@
 #include "pango-modules.h"
 #include "pango-utils.h"
 
-#include FT_TRUETYPE_TABLES_H
+#include <fontconfig/fcfreetype.h>
 
-#define PANGO_SCALE_26_6 (PANGO_SCALE / (1<<6))
-#define PANGO_PIXELS_26_6(d)				\
-  (((d) >= 0) ?						\
-   ((d) + PANGO_SCALE_26_6 / 2) / PANGO_SCALE_26_6 :	\
-   ((d) - PANGO_SCALE_26_6 / 2) / PANGO_SCALE_26_6)
-#define PANGO_UNITS_26_6(d) (PANGO_SCALE_26_6 * (d))
+#include FT_TRUETYPE_TABLES_H
 
 typedef struct _PangoFcMetricsInfo  PangoFcMetricsInfo;
 
@@ -57,6 +52,11 @@ struct _PangoFcFontPrivate
   PangoFcDecoder *decoder;
 };
 
+static gboolean pango_fc_font_real_has_char  (PangoFcFont *font,
+					      gunichar     wc);
+static guint    pango_fc_font_real_get_glyph (PangoFcFont *font,
+					      gunichar     wc);
+
 static void                  pango_fc_font_finalize     (GObject          *object);
 static void                  pango_fc_font_set_property (GObject          *object,
 							 guint             prop_id,
@@ -69,6 +69,7 @@ static PangoCoverage *       pango_fc_font_get_coverage (PangoFont        *font,
 							 PangoLanguage    *language);
 static PangoFontMetrics *    pango_fc_font_get_metrics  (PangoFont        *font,
 							 PangoLanguage    *language);
+static PangoFontMap *        pango_fc_font_get_font_map (PangoFont        *font);
 static PangoFontDescription *pango_fc_font_describe     (PangoFont        *font);
 
 G_DEFINE_TYPE (PangoFcFont, pango_fc_font, PANGO_TYPE_FONT)
@@ -78,6 +79,9 @@ pango_fc_font_class_init (PangoFcFontClass *class)
 {
   GObjectClass *object_class = G_OBJECT_CLASS (class);
   PangoFontClass *font_class = PANGO_FONT_CLASS (class);
+
+  class->has_char = pango_fc_font_real_has_char;
+  class->get_glyph = pango_fc_font_real_get_glyph;
   
   object_class->finalize = pango_fc_font_finalize;
   object_class->set_property = pango_fc_font_set_property;
@@ -85,6 +89,7 @@ pango_fc_font_class_init (PangoFcFontClass *class)
   font_class->find_shaper = pango_fc_font_find_shaper;
   font_class->get_coverage = pango_fc_font_get_coverage;
   font_class->get_metrics = pango_fc_font_get_metrics;
+  font_class->get_font_map = pango_fc_font_get_font_map;
   
   g_object_class_install_property (object_class, PROP_PATTERN,
 				   g_param_spec_pointer ("pattern",
@@ -458,6 +463,45 @@ pango_fc_font_get_metrics (PangoFont     *font,
   return pango_font_metrics_ref (info->metrics);
 }
 
+static PangoFontMap *
+pango_fc_font_get_font_map (PangoFont *font)
+{
+  PangoFcFont *fcfont = PANGO_FC_FONT (font);
+
+  return fcfont->fontmap;
+}
+
+static gboolean
+pango_fc_font_real_has_char (PangoFcFont *font,
+			     gunichar     wc)
+{
+  FcCharSet *charset;
+
+  if (FcPatternGetCharSet (font->font_pattern,
+                           FC_CHARSET, 0, &charset) != FcResultMatch)
+    return FALSE;
+
+  return FcCharSetHasChar (charset, wc);
+}
+
+static guint
+pango_fc_font_real_get_glyph (PangoFcFont *font,
+			      gunichar     wc)
+{
+  FT_Face face;
+  FT_UInt index;
+
+  face = pango_fc_font_lock_face (font);
+  
+  index = FcFreeTypeCharIndex (face, wc);
+  if (index > face->num_glyphs)
+    index = 0;
+
+  pango_fc_font_unlock_face (font);
+
+  return index;
+}
+
 /**
  * pango_fc_font_lock_face:
  * @font: a #PangoFcFont.
