@@ -22,6 +22,8 @@
 /* Size of fontset cache */
 #define FONTSET_CACHE_SIZE 16
 
+#include <math.h>
+
 #include "pango-context.h"
 #include "pangofc-fontmap.h"
 #include "pangofc-private.h"
@@ -61,6 +63,8 @@ struct _PangoFcFontMapPrivate
   PangoFcFamily **families;
   int n_families;		/* -1 == uninitialized */
 
+  double dpi;
+
   /* Decoders */
   GSList *findfuncs;
 
@@ -165,6 +169,7 @@ pango_fc_font_map_init (PangoFcFontMap *fcfontmap)
 					       (GDestroyNotify)g_free,
 					       (GDestroyNotify)pango_coverage_unref);
   priv->fontset_cache = g_queue_new ();
+  priv->dpi = -1;
 }
 
 static void
@@ -233,12 +238,59 @@ pango_fc_coverage_key_equal (PangoFcCoverageKey *key1,
 }
 
 typedef struct _FontsetHashListNode FontsetHashListNode;
+typedef struct _FontsetHashKey      FontsetHashKey;
 
 struct _FontsetHashListNode {
   GHashTable    *fontset_hash;
   PangoLanguage *language;
 };
 
+struct _FontsetHashKey {
+  PangoFontDescription *desc;
+  int x_size;
+  int y_size;
+  guint flags;
+};
+
+static gboolean
+fontset_hash_key_equal (const FontsetHashKey *key_a,
+			const FontsetHashKey *key_b)
+{
+  if (key_a->x_size == key_b->x_size &&
+      key_a->y_size == key_b->y_size &&
+      key_a->flags == key_b->flags &&
+      pango_font_description_equal (key_a->desc, key_b->desc))
+    return TRUE;
+  else
+    return FALSE;
+}
+
+static guint
+fontset_hash_key_hash (const FontsetHashKey *key)
+{
+  return (key->x_size << 16) ^ (key->y_size) ^ (key->flags) ^ pango_font_description_hash (key->desc);
+}
+
+static void
+fontset_hash_key_free (FontsetHashKey *key)
+{
+  pango_font_description_free (key->desc);
+  g_free (key);
+}
+
+static FontsetHashKey *
+fontset_hash_key_copy (FontsetHashKey *old)
+{
+  FontsetHashKey *key = g_new (FontsetHashKey, 1);
+  
+  key->x_size = old->x_size;
+  key->y_size = old->y_size;
+  key->flags = old->flags;
+  key->desc = pango_font_description_copy (old->desc);
+
+  return key;
+}
+
 /* Get the description => fontset map for a particular
  * language tag.
  */
@@ -280,9 +332,9 @@ pango_fc_get_fontset_hash (PangoFcFontMap *fcfontmap,
     priv->fontset_hash_list = g_list_prepend (priv->fontset_hash_list, node);
     
     node->fontset_hash =
-      g_hash_table_new_full ((GHashFunc)pango_font_description_hash,
-			     (GEqualFunc)pango_font_description_equal,
-			     (GDestroyNotify)pango_font_description_free,
+      g_hash_table_new_full ((GHashFunc)fontset_hash_key_hash,
+			     (GEqualFunc)fontset_hash_key_equal,
+			     (GDestroyNotify)fontset_hash_key_free,
 			     (GDestroyNotify)pango_fc_pattern_set_free);
     node->language = language;
 
@@ -592,7 +644,9 @@ pango_fc_convert_width_to_fc (PangoStretch pango_stretch)
 #endif
 
 static FcPattern *
-pango_fc_make_pattern (const PangoFontDescription *description)
+pango_fc_make_pattern (const  PangoFontDescription *description,
+		       PangoLanguage               *language,
+		       double                       pixel_size)
 {
   FcPattern *pattern;
   int slant;
@@ -620,7 +674,7 @@ pango_fc_make_pattern (const PangoFontDescription *description)
 #ifdef FC_WIDTH
 			    FC_WIDTH,  FcTypeInteger, width,
 #endif
- 			    (size_is_absolute ? FC_PIXEL_SIZE : FC_SIZE),  FcTypeDouble,  size,
+ 			    FC_PIXEL_SIZE,  FcTypeDouble,  pixel_size,
 			    NULL);
 
   families = g_strsplit (pango_font_description_get_family (description), ",", -1);
@@ -630,14 +684,20 @@ pango_fc_make_pattern (const PangoFontDescription *description)
 
   g_strfreev (families);
 
+  if (language)
+    FcPatternAddString (pattern, FC_LANG, (FcChar8 *) pango_language_to_string (language));
+  
   return pattern;
 }
 
 static PangoFont *
-pango_fc_font_map_new_font (PangoFontMap      *fontmap,
-			    const PangoMatrix *pango_matrix,
-			    FcPattern     *match)
+pango_fc_font_map_new_font (PangoFontMap               *fontmap,
+			    PangoContext               *context,
+			    const PangoFontDescription *description,
+			    FcPattern                  *match,
+			    gboolean                    cache)
 {
+  PangoFcFontMapClass *class;
   PangoFcFontMap *fcfontmap = (PangoFcFontMap *)fontmap;
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
   FcPattern *pattern;
@@ -654,7 +714,7 @@ pango_fc_font_map_new_font (PangoFontMap      *fontmap,
     return NULL;
   
   /* Look up cache */
-  if (!pango_matrix)
+  if (cache)
     {
       fcfont = g_hash_table_lookup (priv->fonts, match);
   
@@ -662,31 +722,50 @@ pango_fc_font_map_new_font (PangoFontMap      *fontmap,
 	return g_object_ref (fcfont);
     }
 
-  if (pango_matrix)
-    {
-      FcMatrix fc_matrix;
+  class = PANGO_FC_FONT_MAP_GET_CLASS (fontmap);
 
-      /* FontConfig has the Y axis pointing up, Pango, down.
-       */
-      fc_matrix.xx = pango_matrix->xx;
-      fc_matrix.xy = - pango_matrix->xy;
-      fc_matrix.yx = - pango_matrix->yx;
-      fc_matrix.yy = pango_matrix->yy;
-      
-      pattern = FcPatternDuplicate (match);
-      FcPatternAddMatrix (pattern, FC_MATRIX, &fc_matrix);
+  if (class->create_font)
+    {
+      fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->create_font (fcfontmap,
+								   context,
+								   description,
+								   match);
     }
   else
-    pattern = match;
+    {
+      const PangoMatrix *pango_matrix;
 
-  fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, pattern);
+      if (context)
+	pango_matrix = pango_context_get_matrix (context);
+      else
+	pango_matrix = NULL;
 
-  if (!pango_matrix)
+      if (pango_matrix)
+	{
+	  FcMatrix fc_matrix;
+	  
+	  /* FontConfig has the Y axis pointing up, Pango, down.
+	   */
+	  fc_matrix.xx = pango_matrix->xx;
+	  fc_matrix.xy = - pango_matrix->xy;
+	  fc_matrix.yx = - pango_matrix->yx;
+	  fc_matrix.yy = pango_matrix->yy;
+	  
+	  pattern = FcPatternDuplicate (match);
+	  FcPatternAddMatrix (pattern, FC_MATRIX, &fc_matrix);
+	}
+      else
+	pattern = match;
+      
+      fcfont = PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->new_font (fcfontmap, pattern);
+
+      if (pango_matrix)
+	FcPatternDestroy (pattern);
+    }
+
+  if (cache)
     pango_fc_font_map_add (fcfontmap, fcfont);
 
-  if (pango_matrix)
-    FcPatternDestroy (pattern);
-
   fcfont->fontmap = g_object_ref (fcfontmap);
 
   /*
@@ -739,17 +818,111 @@ uniquify_pattern (PangoFcFontMap *fcfontmap,
 
 static void
 pango_fc_default_substitute (PangoFcFontMap    *fontmap,
-			     FcPattern          *pattern)
+			     PangoContext      *context,
+			     FcPattern         *pattern)
 {
-  if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute)
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->context_substitute)
+    PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->context_substitute (fontmap, context, pattern);
+  else if (PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute)
     PANGO_FC_FONT_MAP_GET_CLASS (fontmap)->default_substitute (fontmap, pattern);
 }
 
+static gdouble
+pango_fc_font_map_get_dpi (PangoFcFontMap *fcfontmap,
+			   PangoContext   *context)
+{
+  if (fcfontmap->priv->dpi < 0)
+    {
+      FcResult result = FcResultNoMatch;
+      FcPattern *tmp = FcPatternBuild (NULL,
+				       FC_FAMILY, FcTypeString, "Sans",
+				       FC_SIZE,   FcTypeDouble, 10.,
+				       NULL);
+      if (tmp)
+	{
+	  pango_fc_default_substitute (fcfontmap, NULL, tmp);
+	  result = FcPatternGetDouble (tmp, FC_DPI, 0, &fcfontmap->priv->dpi);
+	  FcPatternDestroy (tmp);
+	}
+      
+      if (result != FcResultMatch)
+	{
+	  g_warning ("Error getting DPI from fontconfig, using 72.0");
+	  fcfontmap->priv->dpi = 72.0;
+	}
+    }
+
+  return fcfontmap->priv->dpi;
+}
+
+static double
+transformed_length (const PangoMatrix *matrix,
+		    double             dx,
+		    double             dy)
+{
+  double tx = (dx * matrix->xx + dy * matrix->xy);
+  double ty = (dx * matrix->yx + dy * matrix->yy);
+
+  return sqrt (tx * tx + ty * ty);
+}
+
+static gboolean
+pango_fc_font_map_get_render_key (PangoFcFontMap              *fcfontmap,
+				  PangoContext                *context,
+				  const PangoFontDescription  *desc,
+				  int                         *x_size,
+				  int                         *y_size,
+				  guint                       *flags)
+{
+  if (PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->get_render_key)
+    return PANGO_FC_FONT_MAP_GET_CLASS (fcfontmap)->get_render_key (fcfontmap, context, desc, x_size, y_size, flags);
+  else
+    {
+      double size;
+      const PangoMatrix *matrix;
+      gboolean retval = TRUE;
+      
+      if (pango_font_description_get_size_is_absolute (desc))
+	size = pango_font_description_get_size (desc);
+      else
+	size = pango_fc_font_map_get_dpi (fcfontmap, context) * pango_font_description_get_size (desc) / 72.;
+      
+      if (context)
+	matrix = pango_context_get_matrix (context);
+      else
+	matrix = NULL;
+
+      *x_size = size;
+      *y_size = size;
+      
+      if (matrix)
+	{
+	  double xl = transformed_length (matrix, 1., 0.);
+	  double yl = transformed_length (matrix, 0., 1.);
+
+	  if (xl != 0. && yl != 0.)
+	    {
+	      *x_size = (int)(xl * *x_size + 0.5);
+	      *y_size = (int)(yl * *y_size + 0.5);
+
+	      retval = (matrix->yx / xl < 1 / 65536. && matrix->xy / yl < 1 / 65536.);
+	    }
+	  else
+	    retval = FALSE;
+	}
+
+      *flags = 0;
+
+      return retval;
+    }
+}
+
 static PangoFcPatternSet *
 pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
-				 PangoContext               *context,
-				 const PangoFontDescription *desc,
-				 PangoLanguage              *language)
+				PangoContext               *context,
+				const PangoFontDescription *desc,
+				PangoLanguage              *language,
+				gboolean                   *cache_out)
 {
   PangoFcFontMap *fcfontmap = (PangoFcFontMap *)fontmap;
   FcPattern *pattern, *font_pattern;
@@ -758,20 +931,27 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
   PangoFcPatternSet *patterns;
   FcFontSet *font_patterns;
   GHashTable *fontset_hash;
+  FontsetHashKey key;
+  gboolean cache;
 
   if (!language && context)
     language = pango_context_get_language (context);
   
-  fontset_hash = pango_fc_get_fontset_hash (fcfontmap, language);
-  patterns = g_hash_table_lookup (fontset_hash, desc);
+ fontset_hash = pango_fc_get_fontset_hash (fcfontmap, language);
+  key.desc = pango_font_description_copy_static (desc);
+  pango_font_description_unset_fields (key.desc, PANGO_FONT_MASK_SIZE);
+  
+  cache = pango_fc_font_map_get_render_key (fcfontmap, context, desc, &key.x_size, &key.y_size, &key.flags);
+  if (cache_out)
+    *cache_out = cache;
+  
+  patterns = g_hash_table_lookup (fontset_hash, &key);
 
   if (patterns == NULL)
     {
-      pattern = pango_fc_make_pattern (desc);
-      if (language)
-	FcPatternAddString (pattern, FC_LANG, (FcChar8 *) pango_language_to_string (language));
+      pattern = pango_fc_make_pattern (desc, language, key.y_size / 1024.);
 
-      pango_fc_default_substitute (fcfontmap, pattern);
+      pango_fc_default_substitute (fcfontmap, context, pattern);
       
       font_patterns = FcFontSort (NULL, pattern, FcTrue, NULL, &res);
 
@@ -820,10 +1000,12 @@ pango_fc_font_map_get_patterns (PangoFontMap               *fontmap,
       FcFontSetSortDestroy (font_patterns);
 
       g_hash_table_insert (fontset_hash,
-			   pango_font_description_copy (desc),
+			   fontset_hash_key_copy (&key),
 			   patterns);
     }
 
+  pango_font_description_free (key.desc);
+
   return patterns;
 }
 
@@ -832,15 +1014,14 @@ pango_fc_font_map_load_font (PangoFontMap               *fontmap,
 			     PangoContext               *context,
 			     const PangoFontDescription *description)
 {
-  PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, description, NULL);
+  PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, description, NULL, NULL);
   if (!patterns)
     return NULL;
 
   if (patterns->n_patterns > 0)
     {
-      return pango_fc_font_map_new_font (fontmap,
-					 context ? pango_context_get_matrix (context) : NULL,
-					 patterns->patterns[0]);
+      return pango_fc_font_map_new_font (fontmap, context, description,
+					 patterns->patterns[0], FALSE);
     }
   
   return NULL;
@@ -903,39 +1084,33 @@ pango_fc_font_map_load_fontset (PangoFontMap                 *fontmap,
 				const PangoFontDescription   *desc,
 				PangoLanguage                *language)
 {
-  PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, desc, language);
+  gboolean cache;
+  PangoFcPatternSet *patterns = pango_fc_font_map_get_patterns (fontmap, context, desc, language, &cache);
   PangoFcFontMap *fcfontmap = PANGO_FC_FONT_MAP (fontmap);
   PangoFcFontMapPrivate *priv = fcfontmap->priv;
   PangoFontset *result;
-  const PangoMatrix *matrix;
   int i;
   
   if (!patterns)
     return NULL;
 
-  if (context)
-    matrix = pango_context_get_matrix (context);
-  else
-    matrix = NULL;
-
-  /* We never cache fontsets when a transformation is in place
-   */
-  if (!patterns->fontset || matrix)
+  if (!patterns->fontset || !cache)
     {
       PangoFontsetSimple *simple;
       simple = pango_fontset_simple_new (language);
       result = PANGO_FONTSET (simple);
-      
+
       for (i = 0; i < patterns->n_patterns; i++)
 	{
 	  PangoFont *font;
 
-	  font = pango_fc_font_map_new_font (fontmap, matrix, patterns->patterns[i]);
+	  font = pango_fc_font_map_new_font (fontmap, context, desc,
+					     patterns->patterns[i], cache);
 	  if (font)
 	    pango_fontset_simple_append (simple, font);
 	}
 
-      if (!matrix)
+      if (cache)
 	{
 	  patterns->fontset = PANGO_FONTSET (simple);
 	  g_object_add_weak_pointer (G_OBJECT (patterns->fontset),
@@ -945,7 +1120,7 @@ pango_fc_font_map_load_fontset (PangoFontMap                 *fontmap,
   else
     result = g_object_ref (patterns->fontset);
   
-  if (!matrix &&
+  if (cache &&
       (!patterns->cache_link ||
        patterns->cache_link != priv->fontset_cache->head))
     pango_fc_font_map_cache_fontset (fcfontmap, patterns);
@@ -988,6 +1163,8 @@ pango_fc_font_map_clear_fontset_cache (PangoFcFontMap *fcfontmap)
 void
 pango_fc_font_map_cache_clear (PangoFcFontMap *fcfontmap)
 {
+  fcfontmap->priv->dpi = -1;
+  
   /* Clear the fontset cache first, since any entries
    * in the fontset_cache must also be in the pattern cache.
    */
@@ -1436,17 +1613,8 @@ pango_fc_face_list_sizes (PangoFontFace  *face,
           if (FcPatternGetDouble (fontset->fonts[i], FC_PIXEL_SIZE, 0, &size) == FcResultMatch)
             {
               if (dpi < 0)
-                {
-                  FcPattern *tmp = FcPatternDuplicate (fontset->fonts[i]);
-                  pango_fc_default_substitute (fcface->family->fontmap, tmp);
-                  if (FcPatternGetDouble (tmp, FC_DPI, 0, &dpi) != FcResultMatch)
-                    {
-                      g_warning ("Error getting DPI from fontconfig, using 72.0");
-                      dpi = 72.0;
-                    }
-                  FcPatternDestroy (tmp);
-                }
-
+		dpi = pango_fc_font_map_get_dpi (fcface->family->fontmap, NULL);
+	      
               size_i = (int) (PANGO_SCALE * size * 72.0 / dpi);
               g_array_append_val (size_array, size_i);
             }
