@@ -67,7 +67,19 @@ struct _PangoFcFontMap
  * @new_font: Creates a new #PangoFcFont for the specified
  *  pattern of the appropriate type for this font map. The
  *  @pattern argument must be passed to the "pattern" property
- *  of #PangoFcFont when you call g_object_new()
+ *  of #PangoFcFont when you call g_object_new(). Deprecated
+ *  in favor of @create_font().
+ * @get_render_key: Given a context and font description,
+ *  calculate a "key" of X and Y sizes and a flags word
+ *  that can be used to hash the results of loading a font
+ *  with that information. If %NULL, a default implementation
+ *  is used.
+ * @create_font: Creates a new #PangoFcFont for the specified
+ *  pattern of the appropriate type for this font map using
+ *  information from the context that is passed in. The
+ *  @pattern argument must be passed to the "pattern" property
+ *  of #PangoFcFont when you call g_object_new(). Deprecated
+ *  in favor of @create_font(). If %NULL, new_font() is used.
  *
  * Class structure for #PangoFcFontMap.
  **/
@@ -77,18 +89,31 @@ struct _PangoFcFontMapClass
   PangoFontMapClass parent_class;
 
   /*< public >*/
+  /* Deprecated in favor of context_substitute */
   void         (*default_substitute) (PangoFcFontMap   *fontmap,
 			              FcPattern        *pattern);
+  /* Deprecated in favor of create_font */
   PangoFcFont  *(*new_font)          (PangoFcFontMap  *fontmap,
 			              FcPattern       *pattern);
 
+  gboolean     (*get_render_key)     (PangoFcFontMap             *fcfontmap,
+				      PangoContext               *context,
+				      const PangoFontDescription *desc,
+				      int                        *xsize,
+				      int                        *ysize,
+				      guint                      *flags);
+  void         (*context_substitute) (PangoFcFontMap             *fontmap,
+				      PangoContext               *context,
+			              FcPattern                  *pattern);
+  PangoFcFont  *(*create_font)       (PangoFcFontMap             *fontmap,
+				      PangoContext               *context,
+				      const PangoFontDescription *desc,
+			              FcPattern                  *pattern);
   /*< private >*/
 
   /* Padding for future expansion */
   void (*_pango_reserved1) (void);
   void (*_pango_reserved2) (void);
-  void (*_pango_reserved3) (void);
-  void (*_pango_reserved4) (void);
 };
 
 PangoContext * pango_fc_font_map_create_context (PangoFcFontMap *fcfontmap);
