@@ -49,6 +49,8 @@ struct _PangoXftFontMap
   PangoXftSubstituteFunc substitute_func;
   gpointer substitute_data;
   GDestroyNotify substitute_destroy;
+
+  PangoRenderer *renderer;
 };
 
 struct _PangoXftFontMapClass
@@ -87,6 +89,9 @@ pango_xft_font_map_finalize (GObject *object)
 {
   PangoXftFontMap *xftfontmap = PANGO_XFT_FONT_MAP (object);
   
+  if (xftfontmap->renderer)
+    g_object_unref (xftfontmap->renderer);
+
   fontmaps = g_slist_remove (fontmaps, object);
 
   if (xftfontmap->substitute_destroy)
@@ -208,7 +213,9 @@ pango_xft_get_font_map (Display *display,
  * @screen: the screen number of a screen within @display
  * 
  * Release any resources that have been cached for the
- * combination of @display and @screen.
+ * combination of @display and @screen. Note that when the
+ * X display is closed, resources are released automatically,
+ * without needing to call this function.
  *
  * Since: 1.2
  **/
@@ -323,6 +330,24 @@ pango_xft_get_context (Display *display,
   return pango_fc_font_map_create_context (PANGO_FC_FONT_MAP (fontmap));
 }
 
+/**
+ * _pango_xft_font_map_get_renderer:
+ * @fontmap: a #PangoXftFontmap
+ * 
+ * Gets the singleton PangoXFTRenderer for this fontmap.
+ * 
+ * Return value: 
+ **/
+PangoRenderer *
+_pango_xft_font_map_get_renderer (PangoXftFontMap *xftfontmap)
+{
+  if (!xftfontmap->renderer)
+    xftfontmap->renderer = pango_xft_renderer_new (xftfontmap->display,
+						   xftfontmap->screen);
+
+  return xftfontmap->renderer;
+}
+
 static void
 pango_xft_font_map_default_substitute (PangoFcFontMap *fcfontmap,
 				       FcPattern      *pattern)
